bl_info = {
    "name": "Show Collection/Object/Data Users of the selected Object in the Viewport",
    "author": "1P2D",
    "version": (1, 0, 2),
    "blender": (4, 2, 0),
    "location": "View3D > Object > Viewport Display",
    "description": "Display the number of collection, object, and data users in the 3D viewport",
    "category": "Object Properties > Viewport Display > Show",
}

import bpy
import blf
from bpy_extras.view3d_utils import location_3d_to_region_2d

handle = None

def register():
    global handle
    # Define BoolProperties for displaying data, object users, and Collection Users
    bpy.types.Object.show_data_users = bpy.props.BoolProperty(
        name="Data Users",
        description="Display the number of data users in the viewport",
        default=False
    )
    bpy.types.Object.show_object_users = bpy.props.BoolProperty(
        name="Object Users",
        description="Display the number of object users (linked across scenes) in the viewport",
        default=False
    )
    bpy.types.Object.show_collection_users = bpy.props.BoolProperty(
        name="Collection Users",
        description="Display the number of collection Users of the object (linked across scenes) in the viewport",
        default=False
    )

    # Append the properties to the UI
    bpy.types.OBJECT_PT_display.append(draw_users_properties)

    # Add the draw handler for the viewport overlay
    if handle is None:
        handle = bpy.types.SpaceView3D.draw_handler_add(
            draw_callback_px, (), 'WINDOW', 'POST_PIXEL')

def unregister():
    global handle
    # Remove the custom properties
    del bpy.types.Object.show_data_users
    del bpy.types.Object.show_object_users
    del bpy.types.Object.show_collection_users

    # Remove the UI panel
    bpy.types.OBJECT_PT_display.remove(draw_users_properties)

    # Remove the draw handler
    if handle is not None:
        bpy.types.SpaceView3D.draw_handler_remove(handle, 'WINDOW')
        handle = None

def draw_users_properties(self, context):
    layout = self.layout
    obj = context.object
    if obj is None:
        return
    col = layout.column(align=True)

    # Check if the object has data
    has_data = hasattr(obj, 'data') and obj.data is not None

    # Create a row for "Data Users" and set its enabled state based on 'has_data'
    row = col.row(align=True)
    row.enabled = has_data
    row.prop(obj, "show_data_users", text="Data Users")

    # "Object Users" is always enabled
    col.prop(obj, "show_object_users", text="Object Users")

    # "Collection Users" is always enabled
    col.prop(obj, "show_collection_users", text="Collection Users")

def count_object_users(obj):
    """
    Count the number of users for the object in terms of scenes and collections,
    aligning with Blender's internal user count logic.
    """
    # Get the number of users directly from Blender's data API
    return obj.users, []


def is_object_in_scene(obj, scene):
    """
    Check if the object is included in any collection of the scene.
    Includes master collections ("Scene Collection") to ensure Object Users count is accurate.
    """
    for collection in obj.users_collection:
        if obj.name in collection.all_objects:
            return True
    return False

def get_linked_collection_info(obj):
    """
    Returns a list of tuples containing the name of each collection the object is part of
    and the number of scenes that include that collection.
    Excludes multiple instances of the same "Scene Collection".
    """
    linked_info = []
    seen_collections = set()  # A set to track unique collections by name

    for collection in obj.users_collection:
        # Skip if it's a master collection to avoid multiple counts of "Scene Collection"
        if hasattr(collection, 'is_master') and collection.is_master:
            if "Scene Collection" not in seen_collections:
                # Only add "Scene Collection" once
                scene_count = count_scenes_with_collection(collection)
                linked_info.append(("Scene Collection", scene_count))
                seen_collections.add("Scene Collection")
            continue

        # Add unique collections
        if collection.name not in seen_collections:
            scene_count = count_scenes_with_collection(collection)
            linked_info.append((collection.name, scene_count))
            seen_collections.add(collection.name)

    return linked_info


def count_scenes_with_collection(collection):
    """
    Count how many scenes include the given collection.
    """
    count = 0
    for scene in bpy.data.scenes:
        if is_collection_in_scene(collection, scene):
            count += 1
    return count

def is_collection_in_scene(collection, scene):
    """
    Recursively check if the collection is part of the scene's collection hierarchy.
    """
    def recursive_search(col):
        if col == collection:
            return True
        for child in col.children:
            if recursive_search(child):
                return True
        return False

    return recursive_search(scene.collection)

def draw_callback_px(*args):
    context = bpy.context

    # Ensure we're in a 3D View
    if not isinstance(context.space_data, bpy.types.SpaceView3D):
        return

    # Ensure overlays are enabled
    if not context.space_data.overlay.show_overlays:
        return

    font_id = 0  # Default font ID (matches "Show Name")
    blf.size(font_id, 11)  # Ensure font size matches "Show Name"

    # Access the theme for consistent coloring
    theme = context.preferences.themes[0]
    view_3d_theme = theme.view_3d

    # Define theme colors
    active_color = (*view_3d_theme.object_active, 1.0)
    selected_color = (*view_3d_theme.object_selected, 1.0)
    unselected_color = (*view_3d_theme.wire, 1.0)
    edit_mode_color = (*view_3d_theme.wire_edit, 1.0)

    # X offset for text positioning
    x_offset = 10

    # Find the active 3D View region
    window_region = None
    for area in context.screen.areas:
        if area.type == 'VIEW_3D':
            for region in area.regions:
                if region.type == 'WINDOW':
                    window_region = region
                    break
            if window_region:
                break
    if not window_region:
        return  # Exit if no 3D View is found

    rv3d = area.spaces.active.region_3d

    # Iterate through visible objects
    for obj in context.visible_objects:
        try:
            # Check if any of the display properties are enabled
            if obj.show_data_users or obj.show_object_users or obj.show_collection_users:
                # Get object's world coordinates
                obj_location = obj.matrix_world.translation

                # Convert 3D coordinates to 2D screen space
                coord_2d = location_3d_to_region_2d(window_region, rv3d, obj_location)

                if coord_2d:
                    texts = []

                    # Handle Data Users
                    if obj.show_data_users:
                        if hasattr(obj, 'data') and obj.data:
                            texts.append(f"Data Users: {obj.data.users}")
                        else:
                            texts.append("Data Users: N/A")

                    # Handle Object Users
                    if obj.show_object_users:
                        # Directly use Blender's internal user count
                        object_user_count = obj.users
                        texts.append(f"Object Users: {object_user_count}")

                    # Handle Collection Users
                    if obj.show_collection_users:
                        linked_collections = get_linked_collection_info(obj)
                        if linked_collections:
                            # Create a string listing all Collection Users in the desired format
                            # Format: X ("Collection Name")
                            collection_texts = [f"{count} (\"{name}\")" for name, count in linked_collections]
                            collections_str = ", ".join(collection_texts)
                            texts.append(f"Collection Users: {collections_str}")
                        else:
                            texts.append("Collection Users: 0")

                    if not texts:
                        continue  # Skip if no texts to display

                    # Determine text color based on object state
                    if obj.mode == 'EDIT':
                        color = edit_mode_color
                    elif obj.select_get():
                        if obj == context.view_layer.objects.active:
                            color = active_color
                        else:
                            color = selected_color
                    else:
                        color = unselected_color

                    # Set the text color
                    blf.color(font_id, *color)

                    # Initialize Y offset for text placement
                    y_offset = coord_2d.y

                    # Adjust position if object name is displayed
                    if getattr(obj, 'show_name', False):
                        object_name = obj.name
                        blf.position(font_id, coord_2d.x + x_offset, coord_2d.y, 0)
                        name_width, name_height = blf.dimensions(font_id, object_name)
                        blf.draw(font_id, object_name)
                        y_offset -= name_height + 4
                    else:
                        y_offset = coord_2d.y + 4

                    # Draw each text line with consistent spacing
                    for text in texts:
                        blf.position(font_id, coord_2d.x + x_offset, y_offset, 0)
                        blf.draw(font_id, text)
                        _, text_height = blf.dimensions(font_id, text)
                        y_offset -= (text_height + 3)  # Consistent spacing between lines

        except Exception as e:
            print(f"Error processing object '{obj.name}': {e}")

if __name__ == "__main__":
    register()
